/*
 * Copyright 1999-2006 University of Chicago
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef GLOBUS_STDIO_UI_MODULE_H
#define GLOBUS_STDIO_UI_MODULE_H 1
#include "globus_common.h"

extern globus_module_descriptor_t              globus_i_stdio_ui_module;
#define GLOBUS_STDIO_UI_MODULE (&globus_i_stdio_ui_module)

#endif /* GLOBUS_STDIO_UI_MODULE_H */
